@@ -43,23 +43,32 @@ class FelderaClient:
 
     def __init__(
         self,
-        url: str,
+        url: Optional[str] = None,
         api_key: Optional[str] = None,
         timeout: Optional[float] = None,
-        requests_verify: bool = True,
+        connection_timeout: Optional[float] = None,
+        requests_verify: bool | str = True,
     ) -> None:
         """
-        :param url: The url to Feldera API (ex: https://try.feldera.com)
+        :param url: The url to Feldera API (ex: https://try.feldera.com). If
+            not set, attempts to read from the environment variable
+            `FELDERA_HOST`. Default: `http://localhost:8080`
         :param api_key: The optional API key for Feldera
-        :param timeout: (optional) The amount of time in seconds that the client
-            will wait for a response before timing
-            out.
+        :param timeout: (optional) The amount of time in seconds that the
+            client will wait for a response before timing out.
+        :param connection_timeout: (optional) The amount of time in seconds
+            that the client will wait to establish connection before timing out
         :param requests_verify: The `verify` parameter passed to the requests
-            library. `True` by default.
+            library. `True` by default. To use a self signed certificate, set
+            it to the path to the certificate.
         """
 
         self.config = Config(
-            url, api_key, timeout=timeout, requests_verify=requests_verify
+            url,
+            api_key,
+            timeout=timeout,
+            connection_timeout=connection_timeout,
+            requests_verify=requests_verify,
         )
         self.http = HttpRequests(self.config)
 
@@ -268,6 +277,58 @@ class FelderaClient:
                 if chunk:
                     yield chunk.decode("utf-8")
 
+    def activate_pipeline(
+        self, pipeline_name: str, wait: bool = True, timeout_s: Optional[float] = 300
+    ):
+        """
+
+        :param pipeline_name: The name of the pipeline to activate
+        :param wait: Set True to wait for the pipeline to activate. True by default
+        :param timeout_s: The amount of time in seconds to wait for the pipeline
+            to activate. 300 seconds by default.
+        """
+
+        if timeout_s is None:
+            timeout_s = 300
+
+        self.http.post(
+            path=f"/pipelines/{pipeline_name}/activate",
+        )
+
+        if not wait:
+            return
+
+        start_time = time.monotonic()
+
+        while True:
+            if timeout_s is not None:
+                elapsed = time.monotonic() - start_time
+                if elapsed > timeout_s:
+                    raise TimeoutError(
+                        f"Timed out waiting for pipeline {pipeline_name} to activate"
+                    )
+
+            resp = self.get_pipeline(pipeline_name)
+            status = resp.deployment_status
+
+            if status == "Running":
+                break
+            elif (
+                status == "Stopped"
+                and len(resp.deployment_error or {}) > 0
+                and resp.deployment_desired_status == "Stopped"
+            ):
+                raise RuntimeError(
+                    f"""Unable to ACTIVATE the pipeline.
+Reason: The pipeline is in a STOPPED state due to the following error:
+{resp.deployment_error.get("message", "")}"""
+                )
+
+            logging.debug(
+                "still starting %s, waiting for 100 more milliseconds", pipeline_name
+            )
+            time.sleep(0.1)
+
     def start_pipeline(
         self, pipeline_name: str, wait: bool = True, timeout_s: Optional[float] = 300
     ):
@@ -504,7 +565,7 @@ Reason: The pipeline is in a STOPPED state due to the following error:
         :raises RuntimeError: If there is currently no transaction in progress.
         :raises ValueError: If the provided `transaction_id` does not match the current transaction.
         :raises TimeoutError: If the transaction does not commit within the specified timeout (when `wait` is True).
-        :raises FelderaAPIError: If the pipeline fails to start a transaction.
+        :raises FelderaAPIError: If the pipeline fails to commit a transaction.
         """
 
         # TODO: implement this without using /stats when we have a better pipeline status reporting API.
@@ -772,6 +833,7 @@ Reason: The pipeline is in a STOPPED state due to the following error:
         backpressure: bool = True,
         array: bool = False,
         timeout: Optional[float] = None,
+        case_sensitive: bool = False,
     ):
         """
         Listen for updates to views for pipeline, yields the chunks of data
@@ -787,6 +849,7 @@ Reason: The pipeline is in a STOPPED state due to the following error:
             "json" format, the default value is False
 
         :param timeout: The amount of time in seconds to listen to the stream for
+        :param case_sensitive: True if the table name is case sensitive or a reserved keyword, False by default
         """
 
         params = {
@@ -797,6 +860,8 @@ Reason: The pipeline is in a STOPPED state due to the following error:
         if format == "json":
             params["array"] = _prepare_boolean_input(array)
 
+        table_name = f'"{table_name}"' if case_sensitive else table_name
+
         resp = self.http.post(
             path=f"/pipelines/{pipeline_name}/egress/{table_name}",
             params=params,
@@ -843,6 +908,27 @@ Reason: The pipeline is in a STOPPED state due to the following error:
             if chunk:
                 yield chunk.decode("utf-8")
 
+    def query_as_hash(self, pipeline_name: str, query: str) -> str:
+        """
+        Executes an ad-hoc query on the specified pipeline and returns a hash of the result.
+
+        :param pipeline_name: The name of the pipeline to query.
+        :param query: The SQL query to be executed.
+        :return: A string containing the hash of the query result.
+        """
+        params = {
+            "pipeline_name": pipeline_name,
+            "sql": query,
+            "format": "hash",
+        }
+
+        resp = self.http.get(
+            path=f"/pipelines/{pipeline_name}/query",
+            params=params,
+            stream=False,
+        )
+        return resp
+
     def query_as_parquet(self, pipeline_name: str, query: str, path: str):
         """
         Executes an ad-hoc query on the specified pipeline and saves the result to a parquet file.
@@ -903,7 +989,7 @@ Reason: The pipeline is in a STOPPED state due to the following error:
             stream=True,
         )
 
-        for chunk in resp.iter_lines(chunk_size=50000000):
+        for chunk in resp.iter_lines(chunk_size=1024):
             if chunk:
                 yield json.loads(chunk, parse_float=Decimal)
 
